"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.splat = void 0;
const path_1 = __importDefault(require("path"));
// @ts-expect-error missing types
const choosealicense_list_1 = __importDefault(require("@rexxars/choosealicense-list"));
const git_remote_origin_url_1 = __importDefault(require("git-remote-origin-url"));
const log_1 = __importDefault(require("../util/log"));
const user_1 = require("../util/user");
const errorToUndefined_1 = require("../util/errorToUndefined");
const package_1 = require("../npm/package");
const prompt_1 = require("../util/prompt");
const readme_1 = require("../util/readme");
const files_1 = require("../util/files");
const bannedFields = ['login', 'description', 'projecturl', 'email'];
const preferredLicenses = ['MIT', 'ISC', 'BSD-3-Clause'];
const otherLicenses = Object.keys(choosealicense_list_1.default.list).filter((id) => {
    const license = choosealicense_list_1.default.list[id];
    return (!preferredLicenses.includes(id) &&
        !bannedFields.some((field) => license.body.includes(`[${field}]`)));
});
function splat(options) {
    var _a, _b, _c;
    return __awaiter(this, void 0, void 0, function* () {
        const { basePath, flags, requireUserConfirmation } = options;
        const info = (write, msg, ...args) => write && log_1.default.info(msg, ...args);
        // Gather data
        const pkg = yield (0, package_1.getPackage)(options).catch(errorToUndefined_1.errorToUndefined);
        log_1.default.debug('Plugin has package.json: %s', pkg ? 'yes' : 'no');
        const user = yield (0, user_1.getUserInfo)(options, pkg);
        log_1.default.debug('User information: %o', user);
        const pkgName = (_a = flags.name) !== null && _a !== void 0 ? _a : pkg === null || pkg === void 0 ? void 0 : pkg.name;
        const pluginName = requireUserConfirmation || !pkgName ? yield (0, prompt_1.promptForPackageName)(options, pkgName) : pkgName;
        log_1.default.debug('Plugin name: %s', pluginName);
        const license = yield getLicense(flags, { user, pluginName, pkg, requireUserConfirmation });
        const licenseChanged = (pkg && pkg.license) !== (license && license.id);
        log_1.default.debug('License: %s', license ? license.id : '<none>');
        const description = yield getProjectDescription(basePath, pkg, requireUserConfirmation);
        log_1.default.debug('Description: %s', description || '<none>');
        const repoUrl = (_b = flags.repo) !== null && _b !== void 0 ? _b : ((yield (0, git_remote_origin_url_1.default)(basePath).catch(errorToUndefined_1.errorToUndefined)) || ((_c = pkg === null || pkg === void 0 ? void 0 : pkg.repository) === null || _c === void 0 ? void 0 : _c.url));
        const gitOrigin = requireUserConfirmation ? yield (0, prompt_1.promptForRepoOrigin)(options, repoUrl) : repoUrl;
        log_1.default.debug('Remote origin: %s', gitOrigin || '<none>');
        // Output
        const data = { user, pluginName, license, description, pkg, gitOrigin };
        let didWrite;
        // Write package.json, if returns the original (data.pkg) if it was unchanged,
        // otherwise it returns the new object
        const newPkg = yield (0, package_1.writePackageJson)(data, options);
        info(newPkg !== pkg, 'Wrote package.json');
        data.pkg = newPkg;
        didWrite = yield writeLicense(data, options, licenseChanged);
        info(didWrite, 'Wrote license file (LICENSE)');
        didWrite = yield writeReadme(data, options);
        info(didWrite, 'Wrote readme file (README.md)');
        didWrite = yield writeStaticAssets(options);
        info(didWrite.length > 0, 'Wrote static asset files: %s', didWrite.join(', '));
        didWrite = yield writeEslintrc(options);
        info(didWrite, 'Wrote .eslintrc.js');
        didWrite = yield (0, package_1.addBuildScripts)(newPkg, options);
        info(didWrite, 'Added build scripts to package.json');
        didWrite = yield addCompileDirToGitIgnore(data, options);
        info(didWrite, 'Added compilation output directory to .gitignore');
    });
}
exports.splat = splat;
function writeReadme(data, options) {
    return __awaiter(this, void 0, void 0, function* () {
        const { basePath } = options;
        const readmePath = path_1.default.join(basePath, 'README.md');
        const readme = yield (0, files_1.readFile)(readmePath, 'utf8').catch(errorToUndefined_1.errorToUndefined);
        if (readme && !(0, readme_1.isDefaultGitHubReadme)(readme)) {
            return false;
        }
        yield (0, files_1.writeFileWithOverwritePrompt)(readmePath, (0, readme_1.generateReadme)(data), {
            encoding: 'utf8',
            force: options.flags.force,
        });
        return true;
    });
}
function writeEslintrc(options) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!options.flags.eslint) {
            return false;
        }
        const { basePath } = options;
        const eslintrc = path_1.default.join(basePath, '.eslintrc');
        const config = {
            root: true,
            env: {
                node: true,
                browser: true,
            },
            extends: [
                'sanity',
                options.flags.typescript && 'sanity/typescript',
                'sanity/react',
                'plugin:react-hooks/recommended',
                options.flags.prettier && 'plugin:prettier/recommended',
            ].filter(Boolean),
        };
        const content = JSON.stringify(config, null, 2);
        yield (0, files_1.writeFileWithOverwritePrompt)(eslintrc, content, {
            encoding: 'utf8',
            force: options.flags.force,
        });
        return true;
    });
}
function writeLicense({ license }, options, licenseChanged) {
    return __awaiter(this, void 0, void 0, function* () {
        const { basePath, flags } = options;
        if (flags.license === false || !license) {
            return false;
        }
        // Prefer whatever path the user is currenly using (LICENSE.md or LICENSE)
        const hasLicenseMdFile = yield (0, files_1.fileExists)(path_1.default.join(basePath, 'LICENSE.md'));
        const licensePath = path_1.default.join(basePath, hasLicenseMdFile ? 'LICENSE.md' : 'LICENSE');
        yield (0, files_1.writeFileWithOverwritePrompt)(licensePath, license.text, {
            encoding: 'utf8',
            default: licenseChanged,
            force: flags.force,
        });
        return true;
    });
}
function getLicense(flags, { user, pluginName, pkg, requireUserConfirmation, }) {
    return __awaiter(this, void 0, void 0, function* () {
        const license = yield getLicenseIdentifier(flags, pkg, requireUserConfirmation);
        if (!license) {
            return undefined;
        }
        const text = license.body
            .replace(/\[fullname\]/g, user === null || user === void 0 ? void 0 : user.name)
            .replace(/\[project\]/g, pluginName)
            .replace(/\[year\]/g, new Date().getFullYear());
        return { id: license.id, text };
    });
}
function getLicenseIdentifier(flags, pkg, requireUserConfirmation = false) {
    return __awaiter(this, void 0, void 0, function* () {
        // --no-license
        if (flags.license === false) {
            return null;
        }
        // --license becomes "", --license mit beocomes "mit"
        if (typeof flags.license === 'string') {
            const license = choosealicense_list_1.default.find(`${flags.license}`);
            if (!license) {
                throw new Error(`License "${flags.license}" not found`);
            }
            return license;
        }
        // no --license flag provided, do we have one in package already?
        if (pkg && pkg.license && !requireUserConfirmation) {
            const license = choosealicense_list_1.default.find(`${pkg.license}`);
            if (license) {
                return license;
            }
            // Warn, then prompt the user
            log_1.default.warn(`package.json contains license "${pkg.license}", which is not recognized`);
        }
        const licenseId = yield (0, prompt_1.prompt)('Which license do you want to use?', {
            default: pkg && pkg.license && choosealicense_list_1.default.find(pkg.license) ? pkg.license : preferredLicenses[0],
            choices: [
                prompt_1.prompt.separator(),
                ...preferredLicenses.map((value) => ({ value, name: choosealicense_list_1.default.list[value].title })),
                prompt_1.prompt.separator(),
                ...otherLicenses.map((value) => ({ value, name: choosealicense_list_1.default.list[value].title })),
            ],
        });
        return choosealicense_list_1.default.find(licenseId);
    });
}
function getProjectDescription(basePath, pkg, requireUserConfirmation = false) {
    return __awaiter(this, void 0, void 0, function* () {
        let description = yield resolveProjectDescription(basePath, pkg);
        if (requireUserConfirmation) {
            description = yield (0, prompt_1.prompt)('Plugin description', { default: description || '' });
        }
        return description !== null && description !== void 0 ? description : '';
    });
}
function resolveProjectDescription(basePath, pkg) {
    return __awaiter(this, void 0, void 0, function* () {
        // Try to grab from package.json
        if (pkg && typeof pkg.description === 'string' && pkg.description.length > 5) {
            return pkg.description;
        }
        // Try to grab a project description from a standard GitHub-generated readme
        try {
            const readmePath = path_1.default.join(basePath, 'README.md');
            const readme = yield (0, files_1.readFile)(readmePath, 'utf8');
            const [title, description] = readme.split('\n').filter(Boolean);
            if (!title || !description || !title.match(/^#\s+\w+/)) {
                return null;
            }
            // Naive, but this isn't too important
            const unlinked = description.replace(/\[(.*?)\]\(.*?\)/g, '$1');
            if (/^[^#]/.test(unlinked)) {
                return unlinked;
            }
            return null;
        }
        catch (err) {
            return (0, errorToUndefined_1.errorToUndefined)(err);
        }
    });
}
function writeStaticAssets({ basePath, flags }) {
    return __awaiter(this, void 0, void 0, function* () {
        const assetsDir = yield findAssetsDir();
        const from = (...segments) => path_1.default.join(assetsDir, 'splat', ...segments);
        const to = (...segments) => path_1.default.join(basePath, ...segments);
        const files = [
            { from: 'editorconfig', to: '.editorconfig' },
            { from: 'npmignore', to: '.npmignore' },
            { from: 'sanity.json', to: 'sanity.json' },
            { from: 'v2-incompatible.js.template', to: 'v2-incompatible.js' },
            flags.gitignore && { from: 'gitignore', to: '.gitignore' },
            flags.typescript && { from: 'template-tsconfig.json', to: 'tsconfig.json' },
            flags.prettier && { from: 'prettierrc.js', to: '.prettierrc.js' },
        ].filter((f) => !!f);
        const writes = [];
        for (const file of files) {
            if (yield (0, files_1.copyFileWithOverwritePrompt)(from(file.from), to(file.to), flags)) {
                writes.push(file.to);
            }
        }
        return writes;
    });
}
/**
 * assets dir might be in higher or lower in the dir hierarchy depending on
 * if we run from lib or src
 */
function findAssetsDir() {
    return __awaiter(this, void 0, void 0, function* () {
        let maxBackpaddle = 3;
        let currDir = __dirname;
        let assetsDir = '';
        while (!assetsDir && maxBackpaddle) {
            currDir = path_1.default.join(currDir, '..');
            const assets = path_1.default.join(currDir, 'assets');
            if (yield (0, files_1.fileExists)(assets)) {
                assetsDir = assets;
            }
            else {
                maxBackpaddle--;
            }
        }
        if (!assetsDir) {
            throw new Error('Could not find assets directory!');
        }
        return assetsDir;
    });
}
function addCompileDirToGitIgnore(data, options) {
    var _a, _b, _c;
    return __awaiter(this, void 0, void 0, function* () {
        const gitIgnorePath = path_1.default.join(options.basePath, '.gitignore');
        const gitignore = yield (0, files_1.readFile)(gitIgnorePath, 'utf8').catch(errorToUndefined_1.errorToUndefined);
        if (!gitignore) {
            return false;
        }
        const output = (_b = (_a = data.pkg) === null || _a === void 0 ? void 0 : _a.main) !== null && _b !== void 0 ? _b : (_c = data.pkg) === null || _c === void 0 ? void 0 : _c.module;
        if (!output) {
            return false;
        }
        const ignorePath = output.replace(/^[./]+/, '');
        const ignore = ignorePath.split('/')[0];
        if (!ignore) {
            return false;
        }
        const lines = gitignore.split('\n');
        if (lines.includes(ignore)) {
            return false;
        }
        lines.push('# Compiled plugin', ignore, '\n');
        yield (0, files_1.writeFile)(gitIgnorePath, lines.join('\n'), { encoding: 'utf8' });
        return true;
    });
}
//# sourceMappingURL=splat.js.map