"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const meow_1 = __importDefault(require("meow"));
const log_1 = __importDefault(require("../util/log"));
const init_1 = require("../actions/init");
const files_1 = require("../util/files");
const manager_1 = require("../npm/manager");
const manifest_1 = require("../sanity/manifest");
const prompt_1 = require("../util/prompt");
const constants_1 = require("../constants");
const description = `Initialize a new Sanity plugin`;
const help = `
Usage
  $ ${constants_1.cliName} init [dir] [<args>]

Options
  --no-eslint             Disables ESLint config and dependencies from being added
  --no-prettier           Disables prettier config and dependencies from being added
  --no-typescript         Disables typescript config and dependencies from being added
  --no-license            Disables LICENSE + package.json license field from being added
  --no-editorconfig       Disables .editorconfig from being added
  --no-gitignore          Disables .gitignore from being added
  --no-scripts            Disables scripts from being added to package.json
  --no-install            Disables automatically running package manager install

  --name [package-name]   Use the provided package-name
  --author [name]         Use the provided author
  --repo [url]            Use the provided repo url
  --license [spdx]        Use the license with the given SPDX identifier
  --force                 No promt when overwriting files

Examples
  # Initialize a new plugin in the current directory
  $ ${constants_1.cliName} init

  # Initialize a plugin in the directory ~/my-plugin
  $ ${constants_1.cliName} init ~/my-plugin

  # Don't add eslint or prettier
  $ ${constants_1.cliName} init --no-eslint --no-prettier
`;
function run({ argv }) {
    return __awaiter(this, void 0, void 0, function* () {
        const cli = (0, meow_1.default)(help, { flags: init_1.initFlags, argv, description });
        const basePath = path_1.default.resolve(cli.input[0] || process.cwd());
        const { exists, isRoot } = yield (0, manifest_1.hasSanityJson)(basePath);
        if (exists && isRoot) {
            throw new Error(`sanity.json has a "root" property set to true - are you trying to init into a studio instead of a plugin?`);
        }
        const { v3ConfigFile } = yield (0, manifest_1.findStudioV3Config)(basePath);
        if (v3ConfigFile) {
            throw new Error(`${v3ConfigFile} exsists - are you trying to init into a studio instead of a plugin?`);
        }
        log_1.default.info('Initializing new plugin in "%s"', basePath);
        if (!cli.flags.force &&
            !(yield (0, files_1.isEmptyish)(basePath)) &&
            !(yield (0, prompt_1.prompt)('Directory is not empty, proceed?', { type: 'confirm', default: false }))) {
            log_1.default.error('Directory is not empty. Cancelled.');
            return;
        }
        yield (0, files_1.ensureDir)(basePath);
        yield (0, init_1.init)({ basePath, flags: cli.flags });
        if (cli.flags.install) {
            if (yield (0, manager_1.installDependencies)(yield (0, manager_1.promptForPackageManager)(), { cwd: basePath })) {
                log_1.default.info('Done!');
            }
            else {
                log_1.default.error('Failed to install dependencies, try manually running `npm install`');
            }
        }
        else {
            log_1.default.info('Dependency installation skipped.');
        }
    });
}
exports.default = run;
//# sourceMappingURL=init.js.map