"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.addBuildScripts = exports.writePackageJson = exports.getReferencedPaths = exports.validatePackage = exports.getPackage = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const util_1 = __importDefault(require("util"));
// @ts-expect-error missing types
const github_url_to_object_1 = __importDefault(require("github-url-to-object"));
// @ts-expect-error missing types
const validate_npm_package_name_1 = __importDefault(require("validate-npm-package-name"));
const manifest_1 = require("../sanity/manifest");
const resolveLatestVersions_1 = require("./resolveLatestVersions");
const files_1 = require("../util/files");
const log_1 = __importDefault(require("../util/log"));
const constants_1 = require("../constants");
const validations_1 = require("../actions/verify/validations");
const readFile = util_1.default.promisify(fs_1.default.readFile);
const pathKeys = ['main', 'module', 'browser', 'types', 'typings'];
function getPackage(opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const options = Object.assign({ flags: {} }, opts);
        validateOptions(options);
        const { basePath, validate = true } = options;
        const manifestPath = path_1.default.normalize(path_1.default.join(basePath, 'package.json'));
        let content;
        try {
            content = yield readFile(manifestPath, 'utf8');
        }
        catch (err) {
            if (err.code === 'ENOENT') {
                throw new Error(`No package.json found. package.json is required to publish to npm. Use \`${constants_1.cliName} init\` for a new plugin, or \`npm init\` for an existing one`);
            }
            throw new Error(`Failed to read "${manifestPath}": ${err.message}`);
        }
        let parsed;
        try {
            parsed = JSON.parse(content);
        }
        catch (err) {
            throw new Error(`Error parsing "${manifestPath}": ${err.message}`);
        }
        if (!isObject(parsed)) {
            throw new Error(`Invalid package.json: Root must be an object`);
        }
        if (validate) {
            yield validatePackage(parsed, options);
        }
        return parsed;
    });
}
exports.getPackage = getPackage;
function validatePackage(manifest, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        validateOptions(opts);
        const options = Object.assign({ isPlugin: true }, opts);
        if (options.isPlugin) {
            yield validatePluginPackage(manifest, options);
        }
        validateLockFiles(options);
    });
}
exports.validatePackage = validatePackage;
function validateOptions(opts) {
    const options = opts || {};
    if (!isObject(options)) {
        throw new Error(`Options must be an object`);
    }
    if (typeof options.basePath !== 'string') {
        throw new Error(`"options.basePath" must be a string (path to plugin base path)`);
    }
}
function validatePluginPackage(manifest, options) {
    return __awaiter(this, void 0, void 0, function* () {
        validatePackageName(manifest);
        yield validatePaths(manifest, options);
    });
}
function validatePackageName(manifest) {
    if (typeof manifest.name !== 'string') {
        throw new Error(`Invalid package.json: "name" must be a string`);
    }
    const valid = (0, validate_npm_package_name_1.default)(manifest.name);
    if (!valid.validForNewPackages) {
        throw new Error(`Invalid package.json: "name" is invalid: ${valid.errors.join(', ')}`);
    }
    const isScoped = manifest.name[0] === '@';
    if (!isScoped && !manifest.name.startsWith('sanity-plugin-')) {
        throw new Error(`Invalid package.json: "name" should be prefixed with "sanity-plugin-" (or scoped - @your-company/plugin-name)`);
    }
}
function validatePaths(manifest, options) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const paths = yield (0, manifest_1.getPaths)(Object.assign(Object.assign({}, options), { pluginName: (_a = manifest.name) !== null && _a !== void 0 ? _a : 'unknown', verifySourceParts: false, verifyCompiledParts: false }));
        const abs = (file) => path_1.default.isAbsolute(file) ? file : path_1.default.resolve(path_1.default.join(options.basePath, file));
        const exists = (file) => fs_1.default.existsSync(abs(file));
        const willExist = (file) => paths && (0, files_1.hasSourceEquivalent)(abs(file), paths);
        const withinSourceDir = (file) => (paths === null || paths === void 0 ? void 0 : paths.source) && abs(file).startsWith(paths.source);
        const withinTargetDir = (file) => (paths === null || paths === void 0 ? void 0 : paths.compiled) && abs(file).startsWith(paths.compiled);
        for (const key of pathKeys) {
            if (!(key in manifest)) {
                continue;
            }
            const manifestValue = manifest[key];
            if (typeof manifestValue !== 'string') {
                throw new Error(`Invalid package.json: "${key}" must be a string if defined`);
            }
            // We don't want to reference `./src/MyComponent.js` containing a bunch of JSX and whatnot,
            // instead we want to target `./lib/MyComponent.js` which is the location it'll be compiled to
            if (!((_b = options === null || options === void 0 ? void 0 : options.flags) === null || _b === void 0 ? void 0 : _b.allowSourceTarget) && paths && withinSourceDir(manifestValue)) {
                throw new Error(`Invalid package.json: "${key}" points to file within source (uncompiled) directory. Use --allow-source-target if you really want to do this.`);
            }
            // Does it exist only because it was there prior to compilation?
            // We're clearing the folder on compilation, so we shouldn't allow it
            const fileExists = exists(manifestValue);
            if (fileExists &&
                paths &&
                withinTargetDir(manifestValue) &&
                !(yield willExist(manifestValue))) {
                throw new Error(`Invalid package.json: "${key}" points to file that will not exist after compiling`);
            }
            // If it _doesn't_ exist and it _won't_ exist, then there isn't much point in continuing, is there?
            if (!exists(manifestValue) && !(yield willExist(manifestValue))) {
                if (!paths) {
                    throw new Error(`Invalid package.json: "${key}" points to file that does not exist`);
                }
                const inOutDir = paths.compiled && !abs(manifestValue).startsWith(paths.compiled);
                throw new Error(inOutDir
                    ? `Invalid package.json: "${key}" points to file that does not exist, and "paths" is not configured to compile to this location`
                    : `Invalid package.json: "${key}" points to file that does not exist, and no equivalent is found in source directory`);
            }
        }
    });
}
function isObject(obj) {
    return !Array.isArray(obj) && obj !== null && typeof obj === 'object';
}
function validateLockFiles(options) {
    const npm = fs_1.default.existsSync(path_1.default.join(options.basePath, 'package-lock.json'));
    const yarn = fs_1.default.existsSync(path_1.default.join(options.basePath, 'yarn.lock'));
    if (npm && yarn) {
        throw new Error(`Invalid plugin: contains both package-lock.json and yarn.lock`);
    }
}
function getReferencedPaths(packageJson, basePath) {
    return pathKeys
        .filter((key) => key in packageJson)
        .map((key) => path_1.default.isAbsolute(packageJson[key])
        ? packageJson[key]
        : path_1.default.resolve(basePath, packageJson[key]));
}
exports.getReferencedPaths = getReferencedPaths;
function writePackageJson(data, options) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const { user, pluginName, license, description, pkg: prevPkg, gitOrigin } = data;
        const { peerDependencies: addPeers, dependencies: addDeps, devDependencies: addDevDeps } = options;
        const { basePath, flags } = options;
        const prev = prevPkg || {};
        const usePrettier = flags.prettier !== false;
        const useEslint = flags.eslint !== false;
        const useTypescript = flags.eslint !== false;
        const newDevDependencies = [constants_1.cliName, 'parcel'];
        if (useTypescript) {
            log_1.default.debug('Using TypeScript. Adding to dev dependencies.');
            newDevDependencies.push('typescript');
        }
        if (usePrettier) {
            log_1.default.debug('Using prettier. Adding to dev dependencies.');
            newDevDependencies.push('prettier');
        }
        if (useEslint) {
            log_1.default.debug('Using eslint. Adding to dev dependencies.');
            newDevDependencies.push('eslint', 'eslint-config-sanity', 'eslint-plugin-react', 'eslint-plugin-react-hooks');
            if (usePrettier) {
                newDevDependencies.push('eslint-config-prettier', 'eslint-plugin-prettier');
            }
            if (useTypescript) {
                newDevDependencies.push('@typescript-eslint/eslint-plugin', '@typescript-eslint/parser');
            }
        }
        log_1.default.debug('Resolving latest versions for %s', newDevDependencies.join(', '));
        const dependencies = Object.assign(Object.assign({}, (prev.dependencies || {})), (addDeps || {}));
        const devDependencies = Object.assign(Object.assign(Object.assign({}, (addDevDeps || {})), (prev.devDependencies || {})), (yield (0, resolveLatestVersions_1.resolveLatestVersions)(newDevDependencies)));
        const peerDependencies = Object.assign(Object.assign({}, (prev.peerDependencies || {})), (addPeers || {}));
        const alwaysOnTop = {
            name: pluginName,
            version: (_a = prev.version) !== null && _a !== void 0 ? _a : '1.0.0',
            description: description || '',
            author: (user === null || user === void 0 ? void 0 : user.email) ? `${user.name} <${user.email}>` : user === null || user === void 0 ? void 0 : user.name,
            license: license ? license.id : 'UNLICENSE',
        };
        const manifest = Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign(Object.assign({}, alwaysOnTop), { source: flags.typescript ? './src/index.ts' : './src/index.js', main: './lib/cjs/index.js', module: './lib/esm/index.js', types: './lib/types/index.d.ts', exports: {
                '.': {
                    require: './lib/cjs/index.js',
                    default: './lib/esm/index.js',
                },
            }, files: ['src', 'lib', 'v2-incompatible.js', 'sanity.json'], scripts: {}, repository: {}, engines: {
                node: '>=14.0.0',
            } }), (prev || {})), alwaysOnTop), { dependencies: sortKeys(dependencies), devDependencies: sortKeys(devDependencies), peerDependencies: sortKeys(peerDependencies) }), repoFromOrigin(gitOrigin)), urlsFromOrigin(gitOrigin));
        const differs = JSON.stringify(prev) !== JSON.stringify(manifest);
        log_1.default.debug('Does manifest differ? %s', differs ? 'yes' : 'no');
        if (differs) {
            yield (0, files_1.writeJsonFile)(path_1.default.join(basePath, 'package.json'), manifest);
        }
        return differs ? manifest : prev;
    });
}
exports.writePackageJson = writePackageJson;
function urlsFromOrigin(gitOrigin) {
    const details = (0, github_url_to_object_1.default)(gitOrigin);
    if (!details) {
        return {};
    }
    return {
        bugs: {
            url: `https://github.com/${details.user}/${details.repo}/issues`,
        },
        homepage: `https://github.com/${details.user}/${details.repo}#readme`,
    };
}
function repoFromOrigin(gitOrigin) {
    if (!gitOrigin) {
        return {};
    }
    return {
        repository: {
            type: 'git',
            url: gitOrigin,
        },
    };
}
function withSanityKeywords(keywords = []) {
    const newKeywords = new Set(keywords);
    newKeywords.add('sanity');
    newKeywords.add('sanity-plugin');
    return Array.from(newKeywords);
}
function addScript(cmd, existing) {
    if (existing && existing.includes(cmd)) {
        return existing;
    }
    return existing ? `${existing} && ${cmd}` : cmd;
}
function addBuildScripts(manifest, { basePath, flags }) {
    return __awaiter(this, void 0, void 0, function* () {
        if (!flags.scripts) {
            return false;
        }
        const originalScripts = manifest.scripts || {};
        const scripts = Object.assign({}, originalScripts);
        scripts.clean = addScript(`rimraf lib`, scripts.clean);
        scripts.lint = addScript(`eslint .`, scripts.lint);
        scripts.prebuild = addScript('npm run clean && ' + validations_1.expectedScripts.prebuild, scripts.prebuild);
        scripts.build = addScript(validations_1.expectedScripts.build, scripts.build);
        scripts.watch = addScript(validations_1.expectedScripts.watch, scripts.watch);
        scripts['link-watch'] = addScript(validations_1.expectedScripts['link-watch'], scripts['link-watch']);
        scripts.prepublishOnly = addScript(validations_1.expectedScripts.prepublishOnly, scripts.prepublishOnly);
        const differs = Object.keys(scripts).some((key) => scripts[key] !== originalScripts[key]);
        if (differs) {
            yield (0, files_1.writeJsonFile)(path_1.default.join(basePath, 'package.json'), Object.assign(Object.assign({}, manifest), { scripts }));
        }
        return differs;
    });
}
exports.addBuildScripts = addBuildScripts;
function sortKeys(unordered) {
    return Object.keys(unordered)
        .sort()
        .reduce((obj, key) => {
        // @ts-expect-error this WILL work
        obj[key] = unordered[key];
        return obj;
    }, {});
}
//# sourceMappingURL=package.js.map